// California Individual-based Fish Simulation System. 
// Copyright (C) 1999 Lang Railsback & Assoc.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.

#import <simtoolsgui.h>
#import "ExperSwarm.h"
#import <math.h>
#import <stdlib.h>
#import <ctype.h>
#import <string.h>

@protocol Model

- (id <Zone>) getModelZone;

@end

@protocol Object

- (id <Zone>) getObjectZone;

@end

@implementation ParameterManager



///////////////////////////////////////////////////////////////////////
//
// canWeGoAgain 
//
///////////////////////////////////////////////////////////////////////
- (BOOL) canWeGoAgain {

   return (modelIterator++ != 5);

}

@end


@implementation ExperSwarm

+ createBegin: aZone
{
  ExperSwarm *obj;
  id <ProbeMap> theProbeMap;

  obj = [super createBegin: aZone];

  obj->numExperimentsRun = 0;

  theProbeMap = [EmptyProbeMap createBegin: aZone];
  [theProbeMap setProbedClass: [self class]];
  theProbeMap = [theProbeMap createEnd];

  [theProbeMap addProbe: [probeLibrary getProbeForVariable: "numExperimentsRun"
                                   inClass: [self class]]];

  [probeLibrary setProbeMap: theProbeMap For: [self class]];

  return obj;		// We return the newly created ExperSwarm
}


- createEnd
{
  return [super createEnd];
}



- buildObjects
{


  [super buildObjects];

  // Build the parameter manager

  parameterManager = [ParameterManager create: self];


  [controlPanel setStateStopped];

  // creates the controller for the subSwarm
  subSwarmControl = [ActivityControl createBegin: [self getZone]];
  [subSwarmControl setDisplayName: "SubSwarm Controller"];
  subSwarmControl = [subSwarmControl createEnd];


  return self;
}

- buildActions
{
  [super buildActions];

  dynRunGroup = [[ActionGroup createBegin: self] createEnd];
  testRunGroup = [[ActionGroup createBegin: self] createEnd];

  experSchedule = [Schedule createBegin: self];
  [experSchedule setRepeatInterval: 1];
  experSchedule = [experSchedule createEnd];

  testSchedule = [Schedule createBegin: self];
  [testSchedule setRepeatInterval: 1];
  testSchedule = [testSchedule createEnd];

  [experSchedule at: 0 createActionTo: self	message: M(setupModel)];
  [experSchedule at: 0 createActionTo: self	message: M(buildModel)];
  [experSchedule at: 0 createActionTo: self	message: M(runModel)];
  [experSchedule at: 0 createAction: dynRunGroup];
  [experSchedule at: 0 createActionTo: self     message: M(checkToStop)];
  [experSchedule at: 0 createActionTo: self	message: M(dropModel)];

  [experSchedule at: 0 createActionTo: probeDisplayManager message: M(update)];
  [experSchedule at: 0 createActionTo: actionCache message: M(doTkEvents)];


  return self;
}  

- activateIn: swarmContext
{
  [super activateIn: swarmContext];
  [experSchedule activateIn: self];
  return [self getActivity];
}




///////////////////////////////////////////////////
//
// setupModel
//
///////////////////////////////////////////////////
- setupModel
{


  subSwarm = [TemplateObserverSwarm create: self];

  [subSwarm objectSetup];


  // Let the subSwarm build its objects and actions and activate
  // it in "nil", giving us a new activity. We don't start it here...
  // we will start models from the ExperSwarm schedule.
  
  [controlPanel setStateStopped];

  return self;

}


- buildModel {

  [subSwarm buildObjects];

  [subSwarm buildActions];

  [subSwarm activateIn: nil];


  // this if is an artefact of bad design for the ActivityControl class
  if (![probeLibrary isProbeMapDefinedFor: [ActivityControl class]]) {
    // now that the subSwarm is activated, attach the controller to it
    [subSwarmControl attachToActivity: [subSwarm getActivity]];
    CREATE_ARCHIVED_PROBE_DISPLAY (subSwarmControl);
  }
  else
    [subSwarmControl attachToActivity: [subSwarm getActivity]];

  return self;
}  




////////////////////////////////////////////////////////////
//
// runModel
//
////////////////////////////////////////////////////////////
- runModel
{
  static id <Symbol> activityState;
  BOOL doneRunning=NO;


  // We have built the model and activated it - here is where we run it.
  // When it has terminated, control will return here.

  while (((activityState = [[subSwarm getActivity] run]) != Terminated) &&
	 (activityState != Stopped))
            
                   ;  // just loop

  printf("\nactivityState = %s \n", [activityState getName]);

  if ((doneRunning = [subSwarm areYouFinishedYet]) == YES) {
    
    fprintf(stderr,"\nEXPERIMENT SWARM::: doneRunning = %d\n", doneRunning);
    fflush(stderr);
 
    numExperimentsRun++;               // increment count of models


  }
  else if (doneRunning == NO) {
    // then insert another runModel onto a dynamic schedule between
    // runModel and doStats
    //[dynRunGroup createActionTo: self message: M(runModel)];


    // and stop the sim so the user can fiddle with things
  
    [controlPanel setStateStopped];

  }


  return self;
}

- dropModel
{
  fprintf(stderr,"<<DROPMODEL>>\n");
  fflush(0);

  // The model has finished and we've extracted the data we need from
  // it. We drop the subSwarm's activity, and then drop the subSwarm
  // itself which drops all of the objects built by subSwarm


  [[subSwarm getActivity] drop];
  [subSwarm drop];
  subSwarm = nil;  

  fprintf(stderr,"<<DROPMODEL>>\n");
  fflush(0);
  xprint(globalZone);
  fprintf(stderr,"<<DROPMODEL>>\n");
  fflush(0);





  return self;

}


- checkToStop
{
  // If all the models have run, time to quit!
  if ([parameterManager canWeGoAgain] == NO)
    {
      
      [probeDisplayManager update];
      [actionCache doTkEvents];
      
      printf("\n All the models have run!\n");
      
      [controlPanel setStateQuit];
    }
  
  return self;
}


@end
